/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fBody.h"
#include "fWake.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTag.h"
#include "wQuad4.h"
#include "wGmshExport.h"
#include <iomanip>
#include <fstream>
using namespace tbox;
using namespace fpm;

Body::Body(std::shared_ptr<MshData> _msh, std::string const &id,
           std::vector<std::string> const &teIds, double xF) : Group(_msh, id), Cl(0), Cd(0), Cs(0), Cm(0)
{
    // If wakes are requested, check if they are already available, otherwise create them
    bool hasChanged = false;
    size_t j = 0;
    for (auto teId : teIds)
    {
        std::cout << "Creating wake... " << std::flush;
        try
        {
            wakes.push_back(new Wake(_msh, tag->name + "Wake" + std::to_string(j), tag));
            std::cout << *wakes.back() << " already existing, nothing done." << std::endl;
        }
        catch (const std::out_of_range &)
        {
            // Notify that mesh needs to be saved
            hasChanged = true;
            // Create tags
            Tag *tagp = new Tag(static_cast<int>(msh->ptags.rbegin()->first + 1), tag->name + "Wake" + std::to_string(j), 2);
            Tag *tage = new Tag(static_cast<int>(msh->etags.rbegin()->first + 1), tag->name + "Wake" + std::to_string(j), 2);
            msh->ntags[tag->name + "Wake" + std::to_string(j)] = tagp;
            msh->ptags[static_cast<int>(msh->ptags.rbegin()->first + 1)] = tagp;
            msh->etags[static_cast<int>(msh->etags.rbegin()->first + 1)] = tage;
            // Sort TE nodes (ascending along y-coordinate)
            Group te(_msh, teId);
            std::vector<Node *> teNodes;
            for (auto e : te.tag->elems)
                for (auto n : e->nodes)
                    teNodes.push_back(n);
            std::sort(teNodes.begin(), teNodes.end(), [](Node *a, Node *b) -> bool { return a->pos(1) < b->pos(1); });
            teNodes.erase(std::unique(teNodes.begin(), teNodes.end()), teNodes.end());
            // Translate TE nodes (along x-coordinate)
            std::vector<Node *> wkNodes;
            for (auto n : teNodes)
            {
                // Sanity check
                if (xF <= n->pos(0))
                {
                    std::stringstream err;
                    err << "fpm::Body: zero or negative length wake requested for " << *tag << "!\n";
                    throw std::runtime_error(err.str());
                }
                Node *nN = new Node(msh->nodes.back()->no + 1, Eigen::Vector3d(xF, n->pos(1), n->pos(2)));
                wkNodes.push_back(nN);
                msh->nodes.push_back(nN);
            }
            // Create wake elements
            for (size_t i = 0; i < wkNodes.size() - 1; ++i)
            {
                std::vector<Node *> qnodes = {teNodes[i], wkNodes[i], wkNodes[i + 1], teNodes[i + 1]};
                msh->elems.push_back(new Quad4(msh->elems.back()->no + 1, tagp, tage, tag->elems[0]->parts, qnodes));
            }
            // Create wake
            wakes.push_back(new Wake(_msh, tag->name + "Wake" + std::to_string(j), tag));
            std::cout << *wakes.back() << " created. " << std::flush;
        }
        ++j;
    }
    // Save the mesh if modified
    if (hasChanged)
    {
        GmshExport writer(_msh);
        writer.save(_msh->name);
    }

    // Size load vectors
    cLoadX.resize(tag->elems.size());
    cLoadY.resize(tag->elems.size());
    cLoadZ.resize(tag->elems.size());
}

/**
 * @brief Save global quantities to file
 */
void Body::save(std::string const &name)
{
    // Write to file
    std::cout << "writing file: " << name + ".dat"
              << "... " << std::flush;
    std::ofstream outfile;
    outfile.open(name + ".dat");
    // Header
    outfile << "$Body - " << tag->name << std::endl;
    // Element count
    outfile << "$Elements" << std::endl;
    outfile << tag->elems.size();
    // Aerodynamic coefficients
    outfile << "$Aerodynamic coefficients" << std::endl;
    outfile << "#" << std::fixed
            << std::setw(14) << "Cl"
            << std::setw(15) << "Cd"
            << std::setw(15) << "Cs"
            << std::setw(15) << "Cm"
            << std::endl;
    outfile << std::fixed
            << std::setw(15) << Cl
            << std::setw(15) << Cd
            << std::setw(15) << Cs
            << std::setw(15) << Cm
            << std::endl;
    // Footer
    outfile << std::endl;
    // Close file
    outfile.close();
    std::cout << "done!" << std::endl;
}

void Body::write(std::ostream &out) const
{
    out << "fpm::Body " << *tag << std::endl;
    out << "with";
    for (auto w : wakes)
        out << "\t" << *w;
}
