/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fProblem.h"
#include "fBody.h"
#include "wTag.h"
#include "wElement.h"
#include "wMem.h"
using namespace tbox;
using namespace fpm;

Problem::Problem(std::shared_ptr<MshData> _smsh, double aoa, double aos, double sref, double cref,
                 double xref, double yref, double zref) : smsh(_smsh), alpha(aoa), beta(aos),
                                                          sR(sref), cR(cref)
{
    xR(0) = xref;
    xR(1) = yref;
    xR(2) = zref;
}

/**
 * @brief Add a non-lifting body
 */
void Problem::add(std::shared_ptr<Body> b)
{
    bodies.push_back(b);
}

/**
 * @brief Update the elements memory (Jacobian)
 */
void Problem::updateMem()
{
    // Update surface Jacobian
    for (auto s : bodies)
        for (auto e : s->tag->elems)
            e->getVMem().update(false);
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 */
void Problem::check() const
{
    // Sanity checks
    if (bodies.empty())
        throw std::runtime_error("No lifting bodies provided!\n");

    // Three-dimension problem
    // Surface element type
    for (auto b : bodies)
        for (auto e : b->tag->elems)
            if (e->type() != ELTYPE::QUAD4)
            {
                std::stringstream err;
                err << "FPM solver is only implemented for surface elements of type "
                    << ELTYPE::QUAD4 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
}

void Problem::write(std::ostream &out) const
{
    out << "fpm::Problem definition"
        << "\n\tAngle of attack: " << alpha
        << "\n\tAngle of sideslip: " << beta
        << "\n\tReference area: " << sR
        << "\n\tReference length: " << cR
        << "\n\tReference point: [" << xR.transpose() << "]"
        << std::endl;
    out << "with";
    for (auto b : bodies)
        out << "\t" << *b;
}
